from __future__ import print_function, unicode_literals
from django.contrib.postgres.fields import JSONField

from .manager import JsonRelatedManager
from . import lookups


class RelatedDocField(JSONField):
    def __init__(self, mapping={}, **kwargs):
        self.mapping = mapping
        super(RelatedDocField, self).__init__(**kwargs)

    def from_db_value(self, value, expression, connection, context):
        return JsonRelatedManager(value, self.mapping)

    def get_default(self):
        return JsonRelatedManager(list(), self.mapping)

    def get_prep_value(self, value):
        if value is not None:
            if isinstance(value, list):
                manager = JsonRelatedManager(list(), self.mapping)
                manager.extend(value)
            elif isinstance(value, JsonRelatedManager):
                manager = value
            else:
                raise ValueError

            return super(RelatedDocField, self).get_prep_value(manager.value)
        return value


RelatedDocField.register_lookup(lookups.ContainsModel)



from psycopg2.extras import Json

from django.db.models import Lookup, Model
from .encoder import JsonDocEncoder


class ContainsModel(Lookup):
    lookup_name = 'has_object'
    prepare_rhs = False

    def process_rhs(self, compiler, connection):
        value = self.rhs
        if isinstance(value, Model):
            lookup_key = JsonDocEncoder(value).key()
            return ('%s', [Json([lookup_key])])
        return super(ContainsModel, self).process_rhs(compiler, connection)

    def as_sql(self, qn, connection):
        lhs, lhs_params = self.process_lhs(qn, connection)
        rhs, rhs_params = self.process_rhs(qn, connection)
        params = lhs_params + rhs_params
        return '%s @> %s' % (lhs, rhs), params



from __future__ import unicode_literals
from django.utils.encoding import python_2_unicode_compatible
from django.core.urlresolvers import reverse
from django.urls.exceptions import NoReverseMatch


@python_2_unicode_compatible
class RelatedDocument(object):
    """
    item element keys
     - key: Reference to correspondent structure in the
     - pk: PK of the object
     - label: (optional) extended label of the object
     - url_args: (optional) list of args to be assed into resolve, instead (pk,) will be used

    """

    def __init__(self, object_manager, item):
        self.object_manager = object_manager
        self.item = item

    def __str__(self):
        if 'label' in self.item:
            return "<%s=%s: %s>" % (self.key, self.pk, self.item['label'])
        return "<%s=%s>" % (self.key, self.pk)

    @property
    def key(self):
        return self.item['key']

    @property
    def pk(self):
        return self.item['pk']

    @property
    def label(self):
        try:
            return self.item['label']
        except KeyError:
            return "%s: %s" % (self.key, self.pk)

    @property
    def instance(self):
        """
        Return instance of the related object with database request
        :return:
        """
        cls = self.object_manager.cls_resolve(self.key)
        return cls.objects.get(pk=self.pk)

    @property
    def fake(self):
        """
        Return fake wrapped object with pk only without request database
        :return:
        """
        cls = self.object_manager.cls_resolve(self.key)
        return cls(pk=self.pk)

    @property
    def url(self):
        url_resolve = self.object_manager.url_resolve(self.key)
        if not url_resolve:
            return None
        url_args = self.item.get('url_args', [self.pk])
        return reverse(url_resolve, args=url_args)

    def get_url(self, default=None):
        try:
            return self.url
        except NoReverseMatch:
            return default



from __future__ import unicode_literals
import six
from importlib import import_module

from django.db import models

from .encoder import JsonDocEncoder
from .document import RelatedDocument
from .serializable import JsonDocument


class JsonRelatedManager(object):
    """
    Mapping item structure:
     0 - Class or string to me imported
     1 - url_resolve sting
    """

    def __init__(self, value, mapping):
        self.value = value
        self.mapping = mapping
        self._cached_imports = {}

    def __str__(self):
        return str(self.value)

    def __getitem__(self, key):
        for item in self.value:
            if item['key'] == key:
                return RelatedDocument(self, item)
        raise KeyError

    def __len__(self):
        return len(self.value)

    def __iter__(self):
        for item in self.value:
            yield RelatedDocument(self, item)

    def get(self, key, default=None):
        """
        Return first occur of object
        :param key:
        :return:
        """
        try:
            return self[key]
        except KeyError:
            return default

    def get_list(self, key):
        rc = []
        for item in self.value:
            if item['key'] == key:
                rc.append(RelatedDocument(self, item))
        return rc

    def cls_resolve(self, key):
        """
        Return class by key
        :param key:
        :return:
        """
        cls_ref = self.mapping[key][0]
        if isinstance(cls_ref, six.string_types):
            if cls_ref not in self._cached_imports:
                module_path = cls_ref.split('.')
                module = import_module('.'.join(module_path[:-1]))
                self._cached_imports[cls_ref] = getattr(module, module_path[-1])
            return self._cached_imports[cls_ref]
        return cls_ref

    def url_resolve(self, key):
        try:
            return self.mapping[key][1]
        except IndexError:
            return None

    @staticmethod
    def _prepare(value):
        if isinstance(value, (models.Model, JsonDocument)):
            enc = JsonDocEncoder(value)
            return enc.dump()
        return value

    def append(self, obj):
        """
        :param obj: either dict or django.db.models.Model instance
        :return:
        """
        self.value.append(self._prepare(obj))

    def extend(self, objects):
        for item in objects:
            self.append(item)

    def insert(self, pos, obj):
        self.value.insert(pos, self._prepare(obj))

    def remove(self, obj):
        value = self._prepare(obj)
        for item in self.value:
            if item['key'] == value['key'] and item['pk'] == value['pk']:
                self.value.remove(item)
                break

    def clear(self):
        self.value = list()
